#!/usr/bin/env python3

import os, sys
import subprocess
import argparse
import time


def getopts():
    p = argparse.ArgumentParser()
    p.add_argument('dir', help='directory of the benchmark')
    p.add_argument('--kratos', help='path to the kratos executable',
                   default='kratos')
    p.add_argument('-e', '--engine', action='append',
                   help='kratos verification engines to use')
    ret = p.parse_args()
    if ret.engine is None:
        ret.engine = [
            'simplic3abs',
            'simplic3kind',
            'simplic3avy'
        ]
    if os.path.basename(ret.kratos) != ret.kratos:
        ret.kratos = os.path.abspath(ret.kratos)
    return ret
    

def info(msg, banner=True):
    n = len(msg) + 2
    if banner:
        print('-' * n)
    for line in msg.splitlines():
        print(' %s ' % line)
    if banner:
        print('-' * n)
    sys.stdout.flush()


def get_verification_engine(opts, engine, progname):
    info('model checking with %s' % engine)
    cmd = [
        opts.kratos, '-stage=mc', '-model_checking_engine=' + engine,
        '-apply_slicing=true', '-verbosity=1', '-trans_use_ite=true',
        '-cex_type=patheval', '-cex_output_file=cex-%s.txt' % engine, progname
        ]
    res = subprocess.Popen(cmd, encoding='utf-8', cwd=opts.dir)
    return res


def run_verification(opts, progname='input.k2'):
    start = time.time()
    provers = []
    for engine in opts.engine:
        p = get_verification_engine(opts, engine, progname)
        provers.append((p, engine))

    res = None
    engine = None
    while res is None:
        some_alive = False
        for p, e in provers:
            if p.poll():
                if p.returncode in (10, 20):
                    res = p.returncode
                    engine = e
                    break
            else:
                some_alive = True
        if not some_alive:
            break

    for p, _ in provers:
        if not p.poll():
            p.kill()

    if res == 20:
        with open(os.path.join(opts.dir, 'cex.txt'), 'w') as out:
            with open(os.path.join(opts.dir, 'cex-%s.txt' % engine)) as f:
                out.write(f.read())

    end = time.time()
    msg = []
    if engine is not None:
        msg.append('solution found by engine: %s' % engine)
    msg.append('total time: %.2f seconds' % (end - start))
    if res == 20:
        cex = os.path.join(opts.dir, 'cex.txt')
        if os.path.exists(cex):
            msg.append('counterexample stored in %s' % cex)
            msg.append('result: unsafe')
        else:
            msg.append('result: unsafe')
    elif res == 10:
        msg.append('result: safe')
    else:
        msg.append('failure in model checking!\nresult: unknown')
    info("\n".join(msg))


def main():
    opts = getopts()
    run_verification(opts)


if __name__ == '__main__':
    main()
